/*** analog 1.9beta ***/
/* Please read Readme.html, or http://www.statslab.cam.ac.uk/~sret1/analog/  */

/*** sscanf.c; functions to replace sscanf(), which is far too slow, in
     certain specific cases ***/

#include <MacHeaders.h>
#include <sioux.h>
#include "analhea2.h"

/*** Now the Mac scanning routines ***/
/*
	WebSART log files have special lines to tell use which fields are present
	and in what order. Parse these so we know how to parse the other lines.
	This also provides a reliable indicator of whether the log is a WebStar
	log to begin with.
	
	A sample field name line:
!!LOG_FORMAT DATE TIME RESULT URL FROM TRANSFER_TIME BYTES_SENT USER HOSTNAME REFERER

	The entries consist of TAB delimited fields. A sample line:
12/14/95	19:00:04	OK  	:pages:Downloads.html		72	3176		indy1.indy.net.	http://www.motu.com/	

	We require the following fields at a minimum:
	HOSTNAME, DATE, URL, BYTES_SENT
	We will also read TIME, RESULT, AGENT and REFERER if present.
*/

enum FieldKinds {
	INVALID = 0,
	HOSTNAME_K,
	DATE,
	URL,
	BYTES_SENT,
	TIME,
	RESULT,
	AGENT,
	FROM,
	METHOD,
	PATH_ARGS,
	REFERER_K,
	SEARCH_ARGS,
	TRANSFER_TIME,
	USER,
	LAST_FIELD
	};

static const char *fnames[LAST_FIELD] = {
	"!!LOG_FORMAT",
	"HOSTNAME",
	"DATE",
	"URL",
	"BYTES_SENT",
	"TIME",
	"RESULT",
	"AGENT",
	"FROM",
	"METHOD",
	"PATH_ARGS",
	"REFERER",
	"SEARCH_ARGS",
	"TRANSFER_TIME",
	"USER"
	};

#define FIELDS_MAX		20

static int fields[FIELDS_MAX];
static int field_cnt = 0;
static long good_lines = 0, bad_lines = 0, tab_lines = 0;

extern int total_succ_reqs;
extern flag warnq;

int
sscanf_webstar(char *inputline, char hostn[MAXSTRINGLENGTH], int *date,
	int *monthno, int *year, int *hr, int *min,
	char filename[MAXSTRINGLENGTH],
	char referer[MAXSTRINGLENGTH], char agent[MAXSTRINGLENGTH],
	int *code, char bytestr[16], size_t preflength)
{     /* scanning 'WebStar' format logfile entries */
	extern flag included();           /* in alias.c */
	extern struct include *noexpandhead, *refexpandhead;
	
	register char *cin = inputline;      /* the character we are reading */
	register char *cout;                 /* where we are putting it */
	int i, field;
	int flags;

	flags=0;
	if (*cin=='!' && !memcmp(cin,fnames[0],strlen(fnames[0]))) {
		while (*cin && *cin!=' ') ++cin;
		if (*cin) ++cin;
		field_cnt=0;
		while (*cin && field_cnt<FIELDS_MAX) {
			for (i=1; i<LAST_FIELD && memcmp(cin,fnames[i],strlen(fnames[i])); ++i) ;
			if (i>=LAST_FIELD) i=INVALID;
			fields[field_cnt]=i;
			++field_cnt;
			flags|=(1<<i);
			while (*cin && *cin!=' ') ++cin;
			if (*cin) ++cin;
			}
		if ((flags>>1)&15!=15 && warnq) {
			fprintf(stderr,"WebStar logs require at least the following fields:\n");
			fprintf(stderr,"  HOSTNAME, DATE, URL, and BYTES_SENT.\n");
			}
		return(0);
		}
	else if (!field_cnt) {
		++bad_lines;
		if (bad_lines<20) {
			for (i=0; *cin; ++cin) if (*cin=='\t') ++i;
			if (i>=3) ++tab_lines;
			}
		if (bad_lines==1 && i==5) {
			field_cnt=6;
			fields[0]=DATE;
			fields[1]=TIME;
			fields[2]=RESULT;
			fields[3]=HOSTNAME_K;
			fields[4]=URL;
			fields[5]=BYTES_SENT;
			}
		else return(0);
		}
	*filename=0;
	*referer=0;
	*agent=0;
	*hr=12;
	*min=0;
	*code=299;
	for (field=0; field<field_cnt; ++field) {
		if (*cin=='\0') break;
		switch (fields[field]) {

		case HOSTNAME_K:  /* read in hostname */
			i = 0;
			for (cout = hostn; *cin != '\t' && *cin != '\0' && i < MAXSTRINGLENGTH - 1;
				 cin++) { 
				*cout = *cin;
				cout++;
				i++;
				}
			if (cout>hostn && *(cout-1)=='.') --cout;
			*cout = '\0';
			flags|=(1<<(HOSTNAME_K-1));
			break;
		case DATE:		/* read in month, day, year */
			/* read in month */
			if (!isdigit(*cin)) return(0);
			else *monthno = 10 * (*cin - '0');
			cin++;
			if (!isdigit(*cin)) return(0);
			else *monthno += (*cin - '0') - 1;
			cin++;

			if (*cin != '/') return(0);
			cin++;
			
			/* read in day */
			if (!isdigit(*cin)) return(0);
			else *date = 10 * (*cin - '0');
			cin++;
			if (!isdigit(*cin)) return(0);
			else *date += (*cin - '0');
			cin++;

			if (*cin != '/') return(0);
			cin++;
			
			/* read in year */
			if (!isdigit(*cin)) return(0);
			else *year = 1900 + 10 * (*cin - '0');
			cin++;
			if (!isdigit(*cin)) return(0);
			else *year += (*cin - '0');
			cin++;
			if (*year < 1970) *year += 100;

			flags|=(1<<(DATE-1));
			break;
		case TIME:		/* read in hour, minute, skip sec */
			/* read in hour */
			if (!isdigit(*cin)) return(0);
			else *hr = 10 * (*cin - '0');
			cin++;
			if (!isdigit(*cin)) return(0);
			else *hr += (*cin - '0');
			cin++;

			if (*cin != ':') return(0);
			cin++;
			
			/* read in minute */
			if (!isdigit(*cin)) return(0);
			else *min = 10 * (*cin - '0');
			cin++;
			if (!isdigit(*cin)) return(0);
			else *min += (*cin - '0');
			cin++;

			if (*cin != ':') return(0);
			cin++;
			
 			/* don't read in second, but check it for correct form */
			if (!isdigit(*cin)) return(0);
			cin++;
			if (!isdigit(*cin)) return(0);
			cin++;

			flags|=(1<<(TIME-1));
			break;
		case URL:		/* read in filename */
			i = 0;
			for (cout = filename; *cin != '\t' && *cin != '\0' && 
				 i < MAXSTRINGLENGTH - 1 - preflength; cin++) {
				if (*cin==':') *cout='/';
				else *cout = tolower(*cin);
				cout++;
				i++;
				}
			*cout = '\0';
			flags|=(1<<(URL-1));
			break;
		case SEARCH_ARGS:/* read in search args (part after '?') */
			if (*filename && !included(filename, noexpandhead)) {  /* read in args */
				cout=filename+strlen(filename);
				*cout++ = '?';
				for ( ; *cin != '\t' && *cin != '\0' &&
					 i < MAXSTRINGLENGTH - 1 - preflength; cin++) {
					*cout = *cin;
					cout++;
					i++;
					}
				*cout = '\0';
				}
			break;
		case BYTES_SENT:/* read in bytes sent */
			i = 0;
			for (cout = bytestr; *cin != '\t' && *cin != '\0' && i < 16; cin++) {
				*cout = *cin;
				cout++;
				}
			*cout = '\0';
			flags|=(1<<(BYTES_SENT-1));
			break;
		case RESULT:	/* Translate WebStar result codes */
			if (cin[0]=='O' && cin[1]=='K') *code=200;
			else if (cin[0]=='E' && cin[1]=='R' && cin[2]=='R' && cin[3]=='!') *code=404;
			else *code=299;
			flags|=(1<<(RESULT-1));
			break;
		case REFERER_K:	/* read in refer */
			i = 0;
			for (cout = referer; *cin != '\0' && *cin != '\t' && *cin != '?' &&
				 i < MAXSTRINGLENGTH - 1; cin++) {
				*cout = *cin;
				cout++;
				i++;
				}
			*cout = '\0';
			if (*cin == '?' && included(referer, refexpandhead)) {  /* read in args */
				for ( ; *cin != '\t' && *cin != '\0' && i < MAXSTRINGLENGTH - 1; cin++) {
					*cout = *cin;
					cout++;
					i++;
					}
				*cout = '\0';
				}
			if (*cin == '?') {
				while (*cin != '\t' && *cin != '\0') cin++;
				}
			flags|=(1<<(REFERER_K-1));
			break;
		case AGENT:		/* read in the user agent */
			i = 0;
			for (cout = agent; *cin != '\0' && *cin != '\t' && i < MAXSTRINGLENGTH - 1;
				 cin++) {
				*cout = *cin;
				cout++;
				i++;
				}
			*cout = '\0';
			flags|=(1<<(AGENT-1));
			break;
			}
		while (*cin && *cin!='\t') ++cin;
		if (*cin) ++cin;
		}
	if ((flags&15)!=15) return(0);
	++good_lines;
	return(11);
	}

void
MacInit(void)
{
	SIOUXSettings.asktosaveonclose=false;
	printf("Processing...\n");
	}

void
MacFini(void)
{
	if (warnq) {
		if (!good_lines && tab_lines>4 && total_succ_reqs<5) {
			fprintf(stderr,"\nWARNING: No valid log entries found!\n");
			fprintf(stderr,"WebStar logs must start with a LOG_FORMAT line something like:\n");
			fprintf(stderr,"!!LOG_FORMAT DATE TIME RESULT URL TRANSFER_TIME BYTES_SENT HOSTNAME REFERER\n");
			fprintf(stderr,"except with the fields listed in the order they actualy occur.\n");
			fprintf(stderr,"WebSTAR 1.2.1 or higher should create this line automaticaly.\n\n");
			}
		}
	else SIOUXSettings.autocloseonquit=true;
	printf("Complete!\n");
	}

void
MacIdle(void)
{
	static long time = 0;

	EventRecord myEvent;
	WindowPtr whichWindow;
	char theChar;

	if (TickCount()<time) return;
	time=TickCount()+6;
	SystemTask();
	if (WaitNextEvent(everyEvent, &myEvent, 1, nil)) {
		if (!SIOUXHandleOneEvent(&myEvent)) switch (myEvent.what) {

		case mouseDown:
			switch (FindWindow(myEvent.where,&whichWindow)) {

			case inMenuBar:
				MenuSelect(myEvent.where);
				break;
			case inSysWindow:
				SystemClick(&myEvent,whichWindow);
				break;
			case inContent:
				SelectWindow(whichWindow);
				break;
			case inDrag:
				DragWindow(whichWindow,myEvent.where,&qd.screenBits.bounds);
				break;
				}
			break;
		case keyDown:
			theChar = myEvent.message & charCodeMask;
			break;
		case updateEvt:
			BeginUpdate((WindowPtr) myEvent.message);
			EndUpdate((WindowPtr) myEvent.message);
			break;
			}
		}
	}

/* end */
